/*
 *  Copyright 2009-2010 Intel Corporation.  All Rights Reserved.
 *
 *  The source code contained or described herein and all documents related
 *  to the source code ("Material") are owned by Intel Corporation or its
 *  suppliers or licensors.  Title to the Material remains with Intel
 *  Corporation or its suppliers and licensors.  The Material is protected
 *  by worldwide copyright laws and treaty provisions.  No part of the
 *  Material may be used, copied, reproduced, modified, published, uploaded,
 *  posted, transmitted, distributed, or disclosed in any way without
 *  Intel's prior express written permission.
 *
 *  No license under any patent, copyright, trade secret or other
 *  intellectual property right is granted to or conferred upon you by
 *  disclosure or delivery of the Materials, either expressly, by
 *  implication, inducement, estoppel or otherwise.  Any license under such
 *  intellectual property rights must be express and approved by Intel in
 *  writing.
 *
 */

#ifndef INCLUDED_CILK_API_UNIX_H
#define INCLUDED_CILK_API_UNIX_H

/* Cilk API -- functions called by the user.  All of these are optional. */

#ifdef IN_CILK_RUNTIME
#pragma GCC visibility push(protected)
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Runtime control.  First argument is parameter name, second value.
 *
 * Allowable names:
 *
 * "nworkers" -- number of processors that should run Cilk code.
 * The value is a number to be parsed by strtol.
 *
 * "force reduce" -- test reducer callbacks by allocating new views
 * for every spawn within which a reducer is accessed.  This can
 * significantly reduce performance.  The value is "1" or "true"
 * to enable, "0" or "false" to disable.
 *
 * Return value is zero on success, nonzero on failure (unrecognized
 * parameter, invalid value, called at the wrong time, or other failure). */
extern __attribute__((nothrow,nonnull(1)))
int __cilkrts_set_param(const char *, const char *);

/* Shut down and deallocate all Cilk state.  The runtime will abort
 * if Cilk is still in use by this thread.  Otherwise the runtime
 * will wait for all other threads using Cilk to exit. */
extern __attribute__((nothrow))
void __cilkrts_end_cilk(void);

/* Allocate Cilk data structures. */
extern __attribute__((nothrow))
void __cilkrts_init(void);

/* get_workers returns the number of worker threads that this instance
 * of Cilk will attempt to use simultaneously.  get_total_workers returns
 * the number of worker descriptors allocated; this number is usually
 * greater than the return value of get_workers.  Both values are constant
 * within a program. */
extern __attribute__((nothrow))
int __cilkrts_get_nworkers(void), __cilkrts_get_total_workers(void);

/* get_worker_number return a small integer indicating which Cilk
 * worker the function is currently running on.  The number is in
 * the range 0..get_total_workers()-1.  */
extern __attribute__((nothrow))
int __cilkrts_get_worker_number(void);

/* Return non-zero if force reduce mode is on */
extern __attribute__((nothrow))
int __cilkrts_get_force_reduce(void);

/* Interact with tools */
extern __attribute__((nothrow))
void __cilkrts_metacall(unsigned int tool, unsigned int code, void *data);

#ifdef __cplusplus
} // extern "C"
#endif

#ifdef IN_CILK_RUNTIME
#pragma GCC visibility pop
#endif

#endif /* INCLUDED_CILK_API_UNIX_H */
